module ForemanAP
  # A virtual machine or container, i.e. a "domain" within libvirt
  #
  class Domain

    require 'rexml/document'

    # Return the hostname of the domain
    def name
      @dom.name
    end

    # Return the amount of memory allocated to the domain, in bytes
    def memory
      # Convert from KiB to bytes
      @dom.info.max_mem.to_i * 1024
    end

    # Return the number of vCPUs allocated to the domain
    def vcpu_count
      @dom.info.nr_virt_cpu
    end

    # Add a storage volume that uses libgfapi.
    #  
    # [+volume_name+] The name of the volume.
    # [+host_name+] The FQDN or IP address of the GlusterFS server.
    # [+device_id+] The position the disk appears on the SCSI bus, starting at one.
    def add_libgfapi_volume(volume_name, host_name, device_id)
      # Determine the target disk name.
      target_dev = 'vd' + ('a'..'z').to_a[device_id - 1]

      disk_xml = REXML::Document.new "
    <disk type='network' device='disk'>
      <driver name='qemu' type='raw' cache='none'/>
      <source protocol='gluster' name='#{volume_name}'>
        <host name='#{host_name}' port='0'/>
      </source>
      <target dev='#{target_dev}' bus='virtio'/>
      <alias name='virtio-disk#{device_id}'/>
    </disk>
      "
      
      # Modify the domain XML to insert the disk
      domain_xml = REXML::Document.new(@dom.xml_desc)
      #puts 'OLD: ' + domain_xml.to_s
      domain_xml.elements.each('domain/devices') do |ele|
        ele.add_element(disk_xml.root)
      end
      #puts 'NEW: ' + domain_xml.to_s

      @dom.undefine
      @dom = @conn.define_domain_xml(domain_xml.to_s)
    end

    # Start (power on) the domain
    def start
      @dom.create
    end

    # Create an object.
    # [+conn+] A connection to libvirtd on a hypervisor
    # [+name+] The name of the domain 
    # [+foreman_api+] A handle to a ForemanAP::ForemanAPI object.
    def initialize(conn, name, foreman_api)
      @conn = conn
      @dom = conn.lookup_domain_by_name(name)
      @foreman_api = foreman_api
    end
  end
end

#--
###### LEGACY STUFF BELOW HERE

class ForemanVM

def snapshot_list
  virsh("snapshot-list #{self.fqdn}")
end
  
# Create a snapshot of the virtual machine
#
def snapshot_create
  raise 'the VM must be powered down before taking a snapshot' \
     if domstate != 'running'
  raise 'a snapshot already exists' if snapshot_list =~ /shutoff/
  virsh("snapshot-create #{fqdn}")
end

# Revert a virtual machine back to a snapshot
#
def snapshot_revert
  raise 'a snapshot does not exist' if snapshot_list =~ /shutoff/
  stop
  virsh("snapshot-revert #{fqdn} --current")
end

# Delete a virtual machine snapshot
#
def snapshot_delete
  virsh("snapshot-delete #{fqdn} --current")
end

# Power off the virtual machine
# 
def stop
  if domstate != 'shut off'
    virsh("destroy #{self.fqdn}")
  end
end


# Get the state of the domain
# 
def domstate
  virsh("domstate #{self.fqdn}").chomp.chomp
end

# Power on the virtual machine
# 
def start
  if domstate != 'running'
    virsh("start #{self.fqdn}")
  end
end


def dumpxml
  puts virsh("dumpxml --security-info #{self.fqdn}")
end

# Modify the VM definition to stop using libgfapi
#
def disable_libgfapi

  self.stop

  require 'rexml/document'
  doc = REXML::Document.new(virsh("dumpxml --security-info #{self.fqdn}"))
  
  # Convert the file-backed disk into a libgfapi disk
  doc.elements.each('domain/devices/disk') do |ele|
   ele.attributes['type'] = 'file'
  end
  doc.elements.each('domain/devices/disk') do |ele|
   ele.delete_element('source')
   ele.add_element('source', {'file'=>"/gvol/images/#{self.fqdn}-disk1", 'protocol' => 'gluster'})
  end
  
  virsh("undefine #{self.fqdn}")
  virsh("define /dev/stdin >/dev/null 2>&1", doc.to_s)
end

# Modify the VM definition to use libgfapi
# +glusterfs_server+ the FQDN of the GlusterFS server
def enable_libgfapi(glusterfs_server)
  require 'rexml/document'
  doc = REXML::Document.new(virsh("dumpxml --security-info #{self.fqdn}"))
  
  raise 'cannot enable libgfapi while the VM is running' if domstate == 'running'
  
  # When cloning or copying, no need to boot from the network
  if @buildspec['_clone'] or @buildspec['_copy']
    doc.delete_element "/domain/os/boot[@dev='network']"
  end

  # Set cache=none just in case
  # Set the disk type, just in case
  doc.elements.each('domain/devices/disk/driver') do |ele|
   ele.attributes['cache'] = 'none'
   ele.attributes['type'] = @buildspec['disk_format']
  end
  
  # Convert the file-backed disk into a libgfapi disk
  doc.elements.each('domain/devices/disk') do |ele|
   ele.attributes['type'] = 'network'
  end
  diskcount = 1  # XXX-KLUDGE: we should actually look at the disk filename
  doc.elements.each('domain/devices/disk') do |ele|
   ele.delete_element('source')
   ele.add_element('source', {'name'=>"gvol/images/#{self.fqdn}-disk#{diskcount}", 'protocol' => 'gluster'})
   diskcount += 1
  end
  doc.elements.each('domain/devices/disk/source') do |ele|
   ele.add_element('host', {'name'=>glusterfs_server, 'transport'=>'tcp', 'port'=>'0'})
  end
  
  virsh("undefine #{self.fqdn}")
  virsh("define /dev/stdin >/dev/null 2>&1", doc.to_s)
end
end
